#include "Template.hpp"

using namespace std;

/*
 * \brief Host Initialization 
 *        Allocate and initialize memory on the host.
 *        Print input array. 
 */
int
initializeHost(void)
{
	width				= /*1280*//*256*/1920;
	height				= /*712*//*256*/1080;
	input				= NULL;
	output				= NULL;
	noPoints			= /*2000*/65536/sizeof(cl_float3);
	/*B 55 L -10
	B 36*/
	top_0=55;//42;//55.0f;
	top_1=-10.0f;
	bottom_0=37;//55;//37.0f;
	bottom_1=20.0f;
	step_0=(bottom_0-top_0)/height;
	step_1=(bottom_1-top_1)/width;

	/////////////////////////////////////////////////////////////////
	// Allocate and initialize memory used by host 
	/////////////////////////////////////////////////////////////////
	cl_uint sizeOutBytes = width*height*sizeof(cl_float);
  output = (cl_float *) malloc(sizeOutBytes);
	if(output == NULL)
	{
		cout<<"Error: Failed to allocate output memory on host\n";
		return 1; 
	}

	ifstream in("awekas_output.txt");
	if(!in.is_open()) {
    cout << "Cannot open awekas_output.txt\n";
    return 1;
  }
	in >> noPoints;
	noPointsCL=noPoints-noPoints%4;
  cout << "noPoints="<<noPoints<<"\n";
  cout << "noPointsCL="<<noPointsCL<<"\n";
  cout << "step_0="<<step_0<<"\n";
  cout << "step_1="<<step_1<<"\n";

	input = (cl_float3 *) malloc(noPoints*sizeof(cl_float3));
	if(input == NULL)
	{
		cout<<"Error: Failed to allocate input memory on host\n";
		return 1; 
	}

  for(int i = 0; i < noPoints; i++) {
		in >> input[i].s[0]/* = width * rand()/(float)RAND_MAX*/;
		in >> input[i].s[1]/* = height * rand()/(float)RAND_MAX*/;
		in >> input[i].s[2]/* = 40.0f * rand()/(float)RAND_MAX*/;
	}


	// print input array
	print1DArray(string("Input").c_str(), input, noPoints);
	return 0;
}

/*
 * Converts the contents of a file into a string
 */
string
convertToString(const char *filename)
{
	size_t size;
	char*  str;
	string s;

	fstream f(filename, (fstream::in | fstream::binary));

	if(f.is_open())
	{
		size_t fileSize;
		f.seekg(0, fstream::end);
		size = fileSize = f.tellg();
		f.seekg(0, fstream::beg);

		str = new char[size+1];
		if(!str)
		{
			f.close();
			return NULL;
		}

		f.read(str, fileSize);
		f.close();
		str[size] = '\0';
	
		s = str;
		delete[] str;
		return s;
	}
	else
	{
		cout << "\nFile containg the kernel code(\".cl\") not found. Please copy the required file in the folder containg the executable.\n";
		exit(1);
	}
	return NULL;
}

/*
 * \brief OpenCL related initialization 
 *        Create Context, Device list, Command Queue
 *        Create OpenCL memory buffer objects
 *        Load CL file, compile, link CL source 
 *		  Build program and kernel objects
 */
int
initializeCL(void)
{
    cl_int status = 0;
    size_t deviceListSize;

    /*
     * Have a look at the available platforms and pick either the AMD one if available or a reasonable default.
     */

    cl_uint numPlatforms;
    cl_platform_id platform = NULL;
    status = clGetPlatformIDs(0, NULL, &numPlatforms);
    if(status != CL_SUCCESS)
    {
        cout << "Error: Getting Platforms. (clGetPlatformsIDs)\n";
        return 1;
    }
    
    if(numPlatforms > 0)
    {
        cl_platform_id* platforms = new cl_platform_id[numPlatforms];
        status = clGetPlatformIDs(numPlatforms, platforms, NULL);
        if(status != CL_SUCCESS)
        {
            cout << "Error: Getting Platform Ids. (clGetPlatformsIDs)\n";
            return 1;
        }
        for(unsigned int i=0; i < numPlatforms; ++i)
        {
            char pbuff[100];
            status = clGetPlatformInfo(
                        platforms[i],
                        CL_PLATFORM_VENDOR,
                        sizeof(pbuff),
                        pbuff,
                        NULL);
            if(status != CL_SUCCESS)
            {
                cout << "Error: Getting Platform Info.(clGetPlatformInfo)\n";
                return 1;
            }
cout << "i="<<i<<"; pbuff="<<pbuff<<"\n";
            platform = platforms[i];
            if(!strcmp(pbuff, "Advanced Micro Devices, Inc."/*"Intel(R) Corporation"*//*"NVIDIA Corporation"*/))
            {
                break;//!
            }
        }
        delete platforms;
    }

    if(NULL == platform)
    {
        cout << "NULL platform found so Exiting Application." << endl;
        return 1;
    }

    /*
     * If we could find our platform, use it. Otherwise use just available platform.
     */
    cl_context_properties cps[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties)platform, 0 };

	/////////////////////////////////////////////////////////////////
	// Create an OpenCL context
	/////////////////////////////////////////////////////////////////
    context = clCreateContextFromType(cps, 
                                      CL_DEVICE_TYPE_GPU,//!
                                      NULL,
                                      NULL,
                                      &status);
    if(status != CL_SUCCESS) 
	{  
		cout<<"Error: Creating Context. (clCreateContextFromType)\n";
		return 1; 
	}

    /* First, get the size of device list data */
    status = clGetContextInfo(context, 
                              CL_CONTEXT_DEVICES, 
                              0, 
                              NULL, 
                              &deviceListSize);
    if(status != CL_SUCCESS) 
	{  
		cout<<
			"Error: Getting Context Info \
		    (device list size, clGetContextInfo)\n";
		return 1;
	}

	/////////////////////////////////////////////////////////////////
	// Detect OpenCL devices
	/////////////////////////////////////////////////////////////////
    devices = (cl_device_id *)malloc(deviceListSize);
	if(devices == 0)
	{
		cout<<"Error: No devices found.\n";
		return 1;
	}

    /* Now, get the device list data */
    status = clGetContextInfo(
			     context, 
                 CL_CONTEXT_DEVICES, 
                 deviceListSize, 
                 devices, 
                 NULL);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
			"Error: Getting Context Info \
		    (device list, clGetContextInfo)\n";
		return 1;
	}

	/////////////////////////////////////////////////////////////////
	// Create an OpenCL command queue
	/////////////////////////////////////////////////////////////////
    commandQueue = clCreateCommandQueue(
					   context, 
                       devices[0], 
                       CL_QUEUE_PROFILING_ENABLE, 
                       &status);
    if(status != CL_SUCCESS) 
	{ 
		cout<<"Creating Command Queue. (clCreateCommandQueue)\n";
		return 1;
	}

	/////////////////////////////////////////////////////////////////
	// Create OpenCL memory buffers
	/////////////////////////////////////////////////////////////////
    inputBuffer = clCreateBuffer(
				      context, 
					  /*CL_MEM_READ_WRITE*/CL_MEM_READ_ONLY | CL_MEM_USE_HOST_PTR,
					  sizeof(cl_float3) * noPoints,
                      input, 
                      &status);
    if(status != CL_SUCCESS) 
	{ 
		cout<<"Error: clCreateBuffer (inputBuffer)\n";
		return 1;
	}

    outputBuffer = clCreateBuffer(
					   context, 
                       /*CL_MEM_READ_WRITE*/CL_MEM_WRITE_ONLY | CL_MEM_USE_HOST_PTR,
					   sizeof(cl_float) * width*height,
                       output, 
                       &status);
    if(status != CL_SUCCESS) 
	{ 
		cout<<"Error: clCreateBuffer (outputBuffer)\n";
		return 1;
	}


	/////////////////////////////////////////////////////////////////
	// Load CL file, build CL program object, create CL kernel object
	/////////////////////////////////////////////////////////////////
    const char * filename  = "Template_Kernels.cl";
    string  sourceStr = convertToString(filename);
    const char * source    = sourceStr.c_str();
    size_t sourceSize[]    = { strlen(source) };

    program = clCreateProgramWithSource(
			      context, 
                  1, 
                  &source,
				  sourceSize,
                  &status);
	if(status != CL_SUCCESS) 
	{ 
	  cout<<
			   "Error: Loading Binary into cl_program \
			   (clCreateProgramWithBinary)\n";
	  return 1;
	}

    /* create a cl program executable for all the devices specified */
    status = clBuildProgram(program, 1, devices, NULL, NULL, NULL);
    if(status != CL_SUCCESS) 
	{ 
		cout<<"Error: Building Program (clBuildProgram)\n";
		return 1; 
	}

    /* get a kernel object handle for a kernel with the given name */
    kernel = clCreateKernel(program, "templateKernel", &status);
    if(status != CL_SUCCESS) 
	{  
		cout<<"Error: Creating Kernel from program. (clCreateKernel)\n";
		return 1;
	}

	return 0;
}


/*
 * \brief Run OpenCL program 
 *		  
 *        Bind host variables to kernel arguments 
 *		  Run the CL kernel
 */
int 
runCLKernels(void)
{
    cl_int   status;
	cl_uint maxDims;
    cl_event events[2];
    size_t globalThreads[2];
//    size_t localThreads[2];
	size_t maxWorkGroupSize;
	size_t maxWorkItemSizes[3];
	/**
	* Query device capabilities. Maximum 
	* work item dimensions and the maximmum
	* work item sizes
	*/ 
	status = clGetDeviceInfo(devices[0], CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(size_t), (void*)&maxWorkGroupSize, NULL);
    if(status != CL_SUCCESS) {  
		cout<<"Error: Getting Device Info. (CL_DEVICE_MAX_WORK_GROUP_SIZE)\n";
		return 1;
	}
	
	status = clGetDeviceInfo(devices[0], CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS, sizeof(cl_uint), (void*)&maxDims, NULL);
    if(status != CL_SUCCESS) {  
		cout<<"Error: Getting Device Info. (CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS)\n";
		return 1;
	}

	status = clGetDeviceInfo(devices[0], CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(size_t)*maxDims, (void*)maxWorkItemSizes, NULL);
    if(status != CL_SUCCESS) {  
		cout<<"Error: Getting Device Info. (CL_DEVICE_MAX_WORK_ITEM_SIZES)\n";
		return 1;
	}
    
cl_uint compute_units;
status = clGetDeviceInfo(devices[0], CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(cl_uint), (void*)&compute_units, NULL);
if(status != CL_SUCCESS) {  
	cout<<"Error: Getting Device Info. (CL_DEVICE_MAX_COMPUTE_UNITS)\n";
	return 1;
}
    
//cl_uint ws = 64;
cl_uint widthPerKernel=1;//920;//!
globalThreads[0] = height;//compute_units * 7 * ws; // 7 wavefronts per SIMD
globalThreads[1] = width/widthPerKernel;//compute_units * 7 * ws; // 7 wavefronts per SIMD
/*while( width % globalThreads[0] != 0 ) globalThreads[0] += ws;
while( height % globalThreads[1] != 0 ) globalThreads[1] += ws;*/
//localThreads[0] = /*ws*/1;

	/*if(globalThreads[0] > maxWorkItemSizes[0]) {
		cout<<"Unsupported: Device ("<<maxWorkItemSizes[0]<<") does not support requested number of global work items ("<<globalThreads[0]<<").";
		return 1;
	}
	if(globalThreads[1] > maxWorkItemSizes[1]) {
		cout<<"Unsupported: Device ("<<maxWorkItemSizes[1]<<") does not support requested number of global work items ("<<globalThreads[1]<<").";
		return 1;
	}*/
	/*if(localThreads[0] > maxWorkGroupSize) {
		cout<<"Unsupported: Device ("<<maxWorkGroupSize<<") does not support requested number of local work items ("<<localThreads[0]<<").";
		return 1;
	}*/

	/*** Set appropriate arguments to the kernel ***/
    /* the output array to the kernel */
    status = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&outputBuffer);
    if(status != CL_SUCCESS) { 
		cout<<"Error: Setting kernel argument. (output)\n";
		return 1;
	}

    /* the input array to the kernel */
    status = clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&inputBuffer);
    if(status != CL_SUCCESS) { 
		cout<<"Error: Setting kernel argument. (input)\n";
		return 1;
	}

    status = clSetKernelArg(kernel, 2, sizeof(cl_uint), (void *)&noPointsCL);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (noPoints)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 3, sizeof(cl_float), (void *)&step_0);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (step_0)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 4, sizeof(cl_float), (void *)&step_1);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (step_1)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 5, sizeof(cl_uint), (void *)&widthPerKernel);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (widthPerKernel)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 6, sizeof(cl_uint), (void *)&width);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (width)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 7, sizeof(cl_float), (void *)&top_0);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (top_0)\n";
		return 1;
	}

	status = clSetKernelArg(kernel, 8, sizeof(cl_float), (void *)&top_1);
    if(status != CL_SUCCESS) { 
		cout<< "Error: Setting kernel argument. (top_1)\n";
		return 1;
	}

		/* 
     * Enqueue a kernel run call.
     */
    status = clEnqueueNDRangeKernel(
			     commandQueue,
                 kernel,
                 /*1*/2,
                 NULL,
                 globalThreads,
                 /*localThreads*/NULL,
                 0,
                 NULL,
                 &events[0]);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
			"Error: Enqueueing kernel onto command queue. \
			(clEnqueueNDRangeKernel)\n";
		return 1;
	}


    /* wait for the kernel call to finish execution */
    status = clWaitForEvents(1, &events[0]);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
		    "Error: Waiting for kernel run to finish. \
			(clWaitForEvents)\n";
		return 1;
	}





cl_ulong time_start, time_end;
double total_time;
 
clGetEventProfilingInfo(events[0], CL_PROFILING_COMMAND_START, sizeof(time_start), &time_start, NULL);
clGetEventProfilingInfo(events[0], CL_PROFILING_COMMAND_END, sizeof(time_end), &time_end, NULL);
total_time = time_end - time_start;
printf("\nExecution time in milliseconds = %0.3f ms\n", (total_time / 1000000.0) );







    status = clReleaseEvent(events[0]);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
		    "Error: Release event object. \
			(clReleaseEvent)\n";
		return 1;
	}

    /* Enqueue readBuffer*/
    status = clEnqueueReadBuffer(
                commandQueue,
                outputBuffer,
                CL_TRUE,
                0,
								width * height* sizeof(cl_float),
                output,
                0,
                NULL,
                &events[1]);
    
    if(status != CL_SUCCESS) 
	{ 
        cout << 
    		"Error: clEnqueueReadBuffer failed. \
             (clEnqueueReadBuffer)\n";

		return 1;
    }
    
    /* Wait for the read buffer to finish execution */
    status = clWaitForEvents(1, &events[1]);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
		    "Error: Waiting for read buffer call to finish. \
			(clWaitForEvents)\n";
		return 1;
	}

    
    status = clReleaseEvent(events[1]);
    if(status != CL_SUCCESS) 
	{ 
		cout<<
		    "Error: Release event object. \
			(clReleaseEvent)\n";
		return 1;
	}

	return 0;
}


/*
 * \brief Release OpenCL resources (Context, Memory etc.) 
 */
int  
cleanupCL(void)
{
    cl_int status;

    status = clReleaseKernel(kernel);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseKernel \n";
		return 1; 
	}
    status = clReleaseProgram(program);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseProgram\n";
		return 1; 
	}
    status = clReleaseMemObject(inputBuffer);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseMemObject (inputBuffer)\n";
		return 1; 
	}
	status = clReleaseMemObject(outputBuffer);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseMemObject (outputBuffer)\n";
		return 1; 
	}
    status = clReleaseCommandQueue(commandQueue);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseCommandQueue\n";
		return 1;
	}
    status = clReleaseContext(context);
    if(status != CL_SUCCESS)
	{
		cout<<"Error: In clReleaseContext\n";
		return 1;
	}

	return 0;
}


/* 
 * \brief Releases program's resources 
 */
void
cleanupHost(void)
{
    if(input != NULL)
    {
        free(input);
        input = NULL;
    }
	if(output != NULL)
	{
		free(output);
		output = NULL;
	}
    if(devices != NULL)
    {
        free(devices);
        devices = NULL;
    }
}


/*
 * \brief Print no more than 256 elements of the given array.
 *
 *        Print Array name followed by elements.
 */
void print1DArray(
		 const string arrayName, 
		 const cl_float3 * arrayData, 
         const unsigned int length)
{
    cl_uint i;
    cl_uint numElementsToPrint = (256 < length) ? 256 : length;

    cout << endl;
    cout << arrayName << ":" << endl;
    for(i = 0; i < numElementsToPrint; ++i)
    {
        cout << arrayData[i].s[0] << "; " << arrayData[i].s[1] << "; " << arrayData[i].s[2] << endl;
    }
    cout << endl;

}

//#define UNDEFINED -1
//#define RETURN_HSV(h, s, v) {HSV.H = h; HSV.S = s; HSV.V = v; return HSV;}
#define RETURN_RGB(r, g, b) {RGB.R = r; RGB.G = g; RGB.B = b; return RGB;}
typedef struct {float R, G, B;} RGBType;
typedef struct {float H, S, V;} HSVType;
RGBType HSV_to_RGB( HSVType HSV ) {
 // H is given on [0, 6] or UNDEFINED. S and V are given on [0, 1].
 // RGB are each returned on [0, 1].
 float h = HSV.H, s = HSV.S, v = HSV.V, m, n, f;
 int i;
 RGBType RGB;

 if (h<0 || h>6) RETURN_RGB(v, v, v);
 i = floor(h);
 f = h - i;
 if ( !(i&1) ) f = 1 - f; // if i is even
 m = v * (1 - s);
 n = v * (1 - s * f);
 switch (i) {
  case 6:
  case 0: RETURN_RGB(v, n, m);
  case 1: RETURN_RGB(n, v, m);
  case 2: RETURN_RGB(m, v, n);
  case 3: RETURN_RGB(m, n, v);
  case 4: RETURN_RGB(n, m, v);
  case 5: RETURN_RGB(v, m, n);
 }
}

/*int RGB_toRGB24(int red, int grn, int blu) {
	return ( ((red) << 16) | ((grn) << 8) | (blu) );
}*/

void savePPM(const cl_float * arrayData) {
	float InverseDistanceSquared_min=0,InverseDistanceSquared_max=37.1f;
	RGBType rgb;
	HSVType hsv;
	char* buf=new char[width*height*3];
	hsv.S=1;
	hsv.V=1;
	// "C:\Program Files (x86)\VideoLAN\VLC\vlc" --demux rawvid --rawvid-fps 1 --rawvid-chroma RV24 --rawvid-width 1920 --rawvid-height 1080 --no-drop-late-frames G:\weather.rv24
	// ffmpeg -vcodec rawvideo -f rawvideo -pix_fmt rgb888 -s 1920x1080 -i weather.rv24 -f image2 -vcodec png weather.png
	ofstream bin ("weather.rv24", ios::out|ios::binary);
	if (bin.is_open()) {
		/*bin.write(string("P6\0a").c_str(),3);
		bin.write(width,2);*/
		for(int y=0,o=0; y<height; y++) {
			for(int x=0; x<width; x++,o+=3) {
				hsv.H=((-InverseDistanceSquared_min+arrayData[y*width+x])/(InverseDistanceSquared_max-InverseDistanceSquared_min))*6.0f;
				if (hsv.H>=0 && hsv.H<=6) {
					HSV_to_RGB(hsv);
					rgb=HSV_to_RGB(hsv);
					buf[o  ]=(char)(rgb.R*255.0f);
					buf[o+1]=(char)(rgb.G*255.0f);
					buf[o+2]=(char)(rgb.B*255.0f);
				}
				/*buf[o  ]=(char)((hsv.H/6.0f)*255.0f);
				buf[o+1]=(char)((hsv.H/6.0f)*255.0f);
				buf[o+2]=(char)((hsv.H/6.0f)*255.0f);*/
				//bin.write(buf,3);
			}
		}
		bin.write(buf,width*height*3);
	} else {
		cout<<"Error: rv24!\n";
	}
}

int 
main(int argc, char * argv[])
{
    // Initialize Host application 
	if(initializeHost()==1)
		return 1;

    // Initialize OpenCL resources
    if(initializeCL()==1)
		return 1;

    // Run the CL program
		//long time_=time(0);
    /*int timer = sampleCommon->createTimer();
    sampleCommon->resetTimer(timer);
    sampleCommon->startTimer(timer);*/
/*		CPerfCounter counter;
		counter.Reset();
		counter.Start();*/
    if(runCLKernels()==1)
		return 1;
		//cout<<time(0)-time_<<endl;
    /*sampleCommon->stopTimer(timer);
		cout<<"totalKernelTime: "<<(double)(sampleCommon->readTimer(timer))<<endl;*/
/*		counter.Stop();
		cout<<"totalKernelTime: "<<counter.GetElapsedTime()<<endl;*/


    // Print output array
		cout<<"savePPM()...\n";
		savePPM(output);
		cout<<"ready!\n";

    // Releases OpenCL resources 
    if(cleanupCL()==1)
		return 1;

    // Release host resources
    cleanupHost();

    return 0;
}
