#ifndef TEMPLATE_H_
#define TEMPLATE_H_


#include <CL/cl.h>
#include <string.h>
#include <cstdlib>
#include <iostream>
#include <string>
#include <fstream>

#include <stdio.h>
//#include <ctime>
#include <math.h>
//#include "Timer.h"

/*** GLOBALS ***/

/*
 * in: cl_float3 points[noPoints]
 * out: cl_float a[width][height]
 *
 * Input data is stored here.
 */
cl_float3 *input;

/*
 * Output data is stored here.
 */
cl_float *output;

/*
 * Multiplier is stored in this variable 
 */
//cl_uint multiplier;

/* problem size for 1D algorithm and width of problem size for 2D algorithm */
cl_uint width, height, noPoints, noPointsCL;
cl_float step_0, step_1, top_0, top_1, bottom_0, bottom_1;

/* The memory buffer that is used as input/output for OpenCL kernel */
cl_mem   inputBuffer;
cl_mem	 outputBuffer;

cl_context          context;
cl_device_id        *devices;
cl_command_queue    commandQueue;

cl_program program;

/* This program uses only one kernel and this serves as a handle to it */
cl_kernel  kernel;


/*** FUNCTION DECLARATIONS ***/
/*
 * OpenCL related initialisations are done here.
 * Context, Device list, Command Queue are set up.
 * Calls are made to set up OpenCL memory buffers that this program uses
 * and to load the programs into memory and get kernel handles.
 */
int initializeCL(void);

/*
 *
 */
std::string convertToString(const char * filename);

/*
 * This is called once the OpenCL context, memory etc. are set up,
 * the program is loaded into memory and the kernel handles are ready.
 * 
 * It sets the values for kernels' arguments and enqueues calls to the kernels
 * on to the command queue and waits till the calls have finished execution.
 *
 * It also gets kernel start and end time if profiling is enabled.
 */
int runCLKernels(void);

/* Releases OpenCL resources (Context, Memory etc.) */
int cleanupCL(void);

/* Releases program's resources */
void cleanupHost(void);

/*
 * Prints no more than 256 elements of the given array.
 * Prints full array if length is less than 256.
 *
 * Prints Array name followed by elements.
 */
void print1DArray(
		 const std::string arrayName, 
         const cl_float3 * arrayData, 
         const unsigned int length);
void savePPM(const cl_float * arrayData, const unsigned int width, const unsigned int height);


#endif  /* #ifndef TEMPLATE_H_ */
